#include "Texture.h"

ioTexture::ioTexture()
{
	loaded = false;
	file = "-";
}

ioTexture::ioTexture( std::string filename )
{
	ioTexture();
	LoadTGA( filename );
}

ioTexture::~ioTexture()
{
	Free();
}

bool ioTexture::LoadTGA( std::string filename )
{
	/*	eby odczyta plik TGA
		potrzebne s odpowiednie
		zmienne pomocnicze.
	*/
	//	Ta tablica jest do porwnania i trzyma nagwek nieskompresowanego pliku TGA
	unsigned char TGAheader[12] = {0,0,2,0,0,0,0,0,0,0,0,0};

	// Waciwa tablica, ktra bdzie trzymac nagwek z pliku
	unsigned char TGAcompare[12];				
	
	// Nagwek opisujcy zapisany obraz
	unsigned char header[6];

	// Zmienna bdzie trzyma ilo bajtw na pixel
	unsigned int bytesPerPixel;				

	// Rozmiar tablicy trzymajcej dane z pliku
	unsigned int imageSize;		

	// Tablica trzymajca dane ( a raczej wskanik na ni )
	unsigned char *imageData;

	// Zmienna pomocnicza
	unsigned int temp;	
	
	// Typ zapisanego obrazu
	unsigned int type = GL_RGBA;

	// Ilo bitw kolorw obrazu
	unsigned int bpp;

	// Jego wysoko i szeroko
	unsigned int width;
	unsigned int height;

	// Otwieramy plik
	FILE *fp = fopen( filename.c_str(), "rb" );

	// Teraz sprawdzamy pare rzeczy
	if(	!fp || // Czy otwarcie pliku si udao
		fread( TGAcompare, 1, sizeof(TGAcompare), fp) != sizeof(TGAcompare) ||	// Czy mona odczyta 12 bajtw nagwka
		memcmp( TGAheader, TGAcompare, sizeof( TGAheader ) ) != 0 ||                    // Czy to jest nagwek ktrego szukamy?
		fread( header, 1, sizeof(header), fp ) != sizeof( header ) )			// Jeeli tak, to czytamy nastpne 6 bajtw
	{
		// Na wypadek gdyby co si nie udao, to zwracamy false
		if ( !fp )
		{
			Log.Error( "TGATEX( " + file + " ): Nieprawidowa cieka lub plik nie istnieje!" );
			return false;
		}
		else
		{
			Log.Error( "TGATEX( " + file + " ): Bd z nagwkiem pliku!" );
			fclose( fp );	
			return false;
		}
	}
	
	if( loaded )
	{
		Log.Report( "TGATEX( " + file + " ): Przeadowanie tekstury na " + filename );

	}
	Log.Log( "TGATEX( " + file + " ): adowanie pliku: " + filename );
	file = filename;

	width  = header[1] * 256 + header[0];		// Opracuj szeroko obrazka
	height = header[3] * 256 + header[2];		// i jego wysokosc
    
 	if(	width  <= 0	||			// Jeeli szeroko
		height <= 0	||			// lub wysokoc jest mniejsza lub rwna zeru
		( header[4] != 24 && header[4] != 32 ) )		// Lub to nie jest ani obraz 24 Bity, ani 32 Bity
	{
		// To zatrzymaj
		Log.Error( "TGATEX( " + file + " ): Nieprawidowy obraz w pliku TGA!" );
		fclose( fp );	
		return false;
	}

	// Odczytaj Bity kolorw obrazu
	bpp = header[4];

	// Oblicz ile to bajtw na pixel
	bytesPerPixel = bpp / 8;

	// Oblicz wielko tablicy
	imageSize = width * height * bytesPerPixel;

	// I zarezerwuj tyle miejsca
	imageData = (unsigned char *)malloc( imageSize );

	/*	Teraz troch teorii
		eby zrobi teksture dla OpenGL
		potrzebna jest tablica z informajcami
		o kolorach. To jest poowa roboty, a pierwsz
		to wydoby te dane z pliku :) Co ju za chwile
		wykonamy.
	*/
	
	if(	imageData == NULL ||	// Sprawd czy pami jest zarezerwowana
		fread( imageData, 1, imageSize, fp ) != imageSize)	// Sprawd czy NA PEWNO odczytalimy tyle z pliku ile chcielimy
	{
		// Jeeli co si nie zgadza, to przerywamy
		if( imageData != NULL)
			free( imageData );

		Log.Error( "TGATEX( " + file + " ): Bd pamici!" );
		fclose( fp );			
		return false;	
	}

	/*	A teraz may kruczek. Jak zapewne
		wiadomo, s trzy kolory skadowe:
		czerwony (Red), zielony (Green) oraz
		niebieski (Blue), czyli poprawny kolotw format powinien
		by RGB. Ale z niewiadomych przyczyn w pliku TGA jest... BGR?
		Musimy to poprawi dlatego teraz zamieniamy
		miejscami dwa feralne bajty.
	*/
	for( unsigned int i = 0; i < int( imageSize ); i += bytesPerPixel )	
	{								
		temp = imageData[i];				// Wkadamy pierwszy bajt ( Blue ) do zmiennej pomocniczej
		imageData[i] = imageData[i + 2];	// Przekadamy ostatni ( Red ) na pierwsze miejsce
		imageData[i + 2] = temp;			// I zastpujemy ostatni wartoci ze zmiennej pomocniczej
	}

	// Zamykamy plik
	fclose( fp );							

	// Musimy jeszcze zmieni format jeeli to nie jest obraz 32 Bity, tylko 24 Bity
	if ( bpp == 24 )
		type = GL_RGB;

	/*	Teraz to dopiero jaja...
		Konwertujemy uzyskane dane na
		teksture(y). Ale dlaczego a 5?
		By mie moliwo wyboru jakoci tekstury :)
		W zalenoci od parametrw jak ma si zachowywa
		tekstura, kiedy jest wiksza, lub mniejsza ni
		w oryginale, to zachowuje rn jako.
		Dodatkowo w najwyszej jakoci uyty jest mipmaping, 
		ktry daje lepszy wygld gdy tekstura jest pod duym ktem.
		Po wicej informacji zapraszam do dokumentacji OpenGL (RTFM!) :) 
	*/

	// Stwrz pi niezdefinowanych tekstur
	glGenTextures( 3, &texture[0] );

	// I same tekstury

	//===========Najgorsza jako, bez MipMapingu==========
	glBindTexture( GL_TEXTURE_2D, texture[0] );			
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	
	glTexImage2D( GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData );

	//===========Najlepsza jako zwyka, bez MipMapingu=====
	glBindTexture( GL_TEXTURE_2D, texture[1] );			
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );  
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );	
	
	glTexImage2D( GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, imageData );

	//===================Najwysza jako MipMapingu====================
	glBindTexture( GL_TEXTURE_2D, texture[2] );			// Bind Our Texture
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR ); 
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	
	gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, imageData );

	// Po wszystkim, te dane s ju niepotrzebne
	delete[] imageData;

	// zaznaczamy, e by odczyt i potwierdzamy poprawne zaadowanie pliku
	loaded = true;
	return true;
}

void ioTexture::Activate( unsigned int tex )
{
	if( !loaded || tex > 2 )
		return;

	glBindTexture( GL_TEXTURE_2D, texture[tex] );
}

void ioTexture::Free()
{
	if( !loaded )
		return;

	glDeleteTextures( 3, &texture[0] );

	loaded = false;
}