#include "glm.h"

/*=====KONSTRUKTOR=====*/
GLModel::GLModel()
{
	// Zerowanie wartoci i przypisywanie wartoci domylnych
	Tex = NULL;
	TexCount = 0;
	List = 0;
	ListCount = 0;
	Frame = 0;
	FrameCount = 0;
	loaded = false;
	file = "-";
}

/*=====DESTRUKTOR=====*/
GLModel::~GLModel()
{
	// Zwolnienie pamici
	Free();
}

/*=====METODA GetString=====
	Metoda pobiera jedn linie z pliku
	i j modyfikuje. Pozbywa si pustej
	linii, znakw spacji i tabulacji przed
	rozpoczciem waciwych polece, oraz
	czyci z komentarzy.
*/
std::string GLModel::GetString( FILE* fp )
{
	char buf[1024];
	std::string zwrot;
	int len = 0;
	bool repeat = false;
	bool wfchar = false;

	do
	{
		len = 0;
		repeat = false;
		wfchar = false;
		fgets( buf, 1024, fp );
	
		len = strlen( buf );
		zwrot = "";

		if( buf[0] == '\n' )
		{
			repeat = true;
			continue;
		}

		for( int i = 0; i < len; i++ )
		{
			if( ( buf[i] == '#' || buf[i] == '/' ) && !wfchar )
			{
				repeat = true;
				break;
			}

			if( buf[i] != ' ' && buf[i] != '\t' && buf[i] != '\n' )
				wfchar = true;

			zwrot += buf[i];
		}
	}
	while( repeat );

	if( zwrot[zwrot.length()-1] == '\n' )
		zwrot = zwrot.substr( 0, zwrot.length() - 1 );

	return zwrot;
}

/*=====METODA NoSpace=====
	Czyci podan lini ze spacji, tablulacji
	i znakw nowej linii.
*/
std::string GLModel::NoSpace( std::string str )
{
	std::string zwrot = "";

	for( int i = 0; i < str.length(); i++ )
	{
		if( str[i] == ' ' || str[i] == '\t' || str[i] == '\n' )
			continue;

		zwrot += str[i];
	}

	return zwrot;
}

/*=====METODA GetParams=====
	Pobiera dowoln ilo parametrw pomidzy
	nawiasami ( ) odzielonymi przecinkiem i je
	wpisuje do podanej tablicy.
*/
bool GLModel::GetParams( std::string str, int from, std::string* param, int paramCount, std::string Com )
{
	if( paramCount <= 0 )
		return true;

	int curparam = 0;
	int i;
	for( i = from; i < str.length(); i++ )
	{
		if( str[i] == ')' )
			break;

		if( str[i] == ',' )
		{
			curparam++;
			i++;

			if( curparam > paramCount )
				break;
		}

		param[curparam] += str[i];
	}
	if( curparam > paramCount )
	{
		Log.Error( "GLMODEL( " + file + " ): Za duo parametrw polecenia: " + Com );
		return false;
	}
	else if( curparam != paramCount-1 )
	{
		Log.Error( "GLMODEL( " + file + " ): Za mao parametrw polecenia: " + Com );	
		return false;
	}

	return true;
}
/*=====METODA GetConst=====
	Metoda zwraca sta z OpenGL, w zalenoci od nazwy
	i nazwy polecenia, do jakiej jest potrzebna.
*/
int GLModel::GetConst( std::string str, std::string Com )
{
	if( Com == "glEnable" || Com == "glDisable" )
	{
		if( str == "GL_DEPTH_TEST" )
			return GL_DEPTH_TEST;

		if( str == "GL_BLEND" )
			return GL_BLEND;

		if( str == "GL_TEXTURE_2D" )
			return GL_TEXTURE_2D;

		if( str == "GL_NORMALIZE" )
			return GL_NORMALIZE;

		if( str == "GL_CULL_FACE" )
			return GL_CULL_FACE;

	}
	if( Com == "glBegin" )
	{
		if( str == "GL_POINTS" )
			return GL_POINTS;

		if( str == "GL_LINES" )
			return GL_LINES;

		if( str == "GL_LINE_STRIP" )
			return GL_LINE_STRIP;

		if( str == "GL_LINE_LOOP" )
			return GL_LINE_LOOP;

		if( str == "GL_TRIANGLES" )
			return GL_TRIANGLES;

		if( str == "GL_TRIANGLE_STRIP" )
			return GL_TRIANGLE_STRIP;

		if( str == "GL_TRIANGLE_FAN" )
			return GL_TRIANGLE_FAN;

		if( str == "GL_QUADS" )
			return GL_QUADS;

		if( str == "GL_QUAD_STRIP" )
			return GL_QUAD_STRIP;

		if( str == "GL_POLYGON" )
			return GL_POLYGON;
	}
	if( Com == "glBlendFunc" )
	{
		if( str == "GL_ZERO" )
			return GL_ZERO;

		if( str == "GL_ONE" )
			return GL_ONE;
		
		if( str == "GL_DST_COLOR" )
			return GL_DST_COLOR;
		
		if( str == "GL_ONE_MINUS_DST_COLOR" )
			return GL_ONE_MINUS_DST_COLOR;
		
		if( str == "GL_SRC_ALPHA" )
			return GL_SRC_ALPHA;
		
		if( str == "GL_ONE_MINUS_SRC_ALPHA" )
			return GL_ONE_MINUS_SRC_ALPHA;
		
		if( str == "GL_DST_ALPHA" )
			return GL_DST_ALPHA;
			
		if( str == "GL_ONE_MINUS_DST_ALPHA" )
			return GL_ONE_MINUS_DST_ALPHA;
		
		if( str == "GL_SRC_ALPHA_SATURATE" )
			return GL_SRC_ALPHA_SATURATE;
	} 
	if( Com == "glFrontFace" )
	{
		if( str == "GL_CW" )
			return GL_CW;

		if( str == "GL_CCW" )
			return GL_CCW;
	}
	if( Com == "gluQuadricDrawStyle" )
	{
		if( str == "GLU_FILL" )
			return GLU_FILL;
		
		if( str == "GLU_LINE" )
			return GLU_LINE;
		
		if( str == "GLU_SILHOUETTE" )
			return GLU_SILHOUETTE;

		if( str == "GLU_POINT" )
			return GLU_POINT;
	}
	if( Com == "gluQuadricNormals" )
	{
		if( str == "GLU_NONE" )
			return GLU_NONE;

		if( str == "GLU_FLAT" )
			return GLU_FLAT;

		if( str == "GLU_SMOOTH" )
			return GLU_SMOOTH;
	}
	if( Com == "gluQuadricOrientation" )
	{
		if( str == "GLU_OUTSIDE" )
			return GLU_OUTSIDE;

		if( str == "GLU_INSIDE" )
			return GLU_INSIDE;
	}
	if( Com == "gluQuadricTexture" )
	{
		if( str == "GL_TRUE" )
			return GL_TRUE;

		if( str == "GL_FALSE" )
			return GL_FALSE;
	}
	Log.Error( "GLMODEL( " + file + " ): Nieznana staa ( " + str + " ) w poleceniu: " + Com );
	return atoi( str.c_str() );
}

/*=====METODA ParseGLCommand=====
	Analizuje linie, szuka nazwy komendy OpenGL, pobiera
	dla niej argumenty i j wykonuje.
*/
void GLModel::ParseGLCommand( std::string str )
{
	str = NoSpace( str );

	std::string param[6];
	bool presult = false;
	std::string Com = "";
	int i = 0, k;

	for( i = 0; i < str.length(); i++ )
	{
		if( str[i] == '(' )
			break;

		Com += str[i];
	}

	if( Com == "glBindTexture" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			k = atoi( param[0].c_str() );
			if( k >= TexCount || k < 0 )
			{
				Log.Error( "GLMODEL( " + file + " ): Bdny parametr polecenia: " + Com );
			}
			else Tex[k].Activate(  );
			return;
		}
	}
	if( Com == "glCallList" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			k = atoi( param[0].c_str() );
			if( k >= ListCount || k < 0 )
			{
				Log.Error( "GLMODEL( " + file + " ): Bdny parametr polecenia: " + Com );
			}
			else glCallList( List + k );
			return;
		}
	}
	if( Com == "glRotate" )
	{
		presult = GetParams( str, i+1, param, 4, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" && param[3] != "" )
		{
			glRotatef( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ), atof( param[3].c_str() ) );
			return;
		}
	}
	if( Com == "glTranslate" )
	{
		presult = GetParams( str, i+1, param, 3, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" )
		{
			glTranslatef( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ) );
			return;
		}
	}
	if( Com == "glScale" )
	{
		presult = GetParams( str, i+1, param, 3, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" )
		{
			glScalef( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ) );
			return;
		}
	}
	if( Com == "glBegin" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			glBegin( GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "glEnable" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			glEnable( GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "glDisable" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			glDisable( GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "glEnd" )
	{
		presult = GetParams( str, i+1, param, 0, Com );

		if( presult )
		{
			glEnd();
			return;
		}
	}
	if( Com == "glPushMatrix" )
	{
		presult = GetParams( str, i+1, param, 0, Com );

		if( presult )
		{
			glPushMatrix();
			return;
		}
	}
	if( Com == "glPopMatrix" )
	{
		presult = GetParams( str, i+1, param, 0, Com );

		if( presult )
		{
			glPopMatrix();
			return;
		}
	}
	if( Com == "glColor3" )
	{
		presult = GetParams( str, i+1, param, 3, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" )
		{
			glColor3f( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ) );
			return;
		}
	}
	if( Com == "glColor4" )
	{
		presult = GetParams( str, i+1, param, 4, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" && param[3] != "" )
		{
			glColor4f( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ), atof( param[3].c_str() ) );
			return;
		}
	}
	if( Com == "glNormal3" )
	{
		presult = GetParams( str, i+1, param, 3, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" )
		{
			glNormal3f( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ) );
			return;
		}
	}
	if( Com == "glTexCoord2" )
	{
		presult = GetParams( str, i+1, param, 2, Com );

		if( presult && param[0] != "" && param[1] != "" )
		{
			glTexCoord2f( atof( param[0].c_str() ), atof( param[1].c_str() ) );
			return;
		}
	}
	if( Com == "glVertex3" )
	{
		presult = GetParams( str, i+1, param, 3, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" )
		{
			glVertex3f( atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ) );
			return;
		}
	}
	if( Com == "glBlendFunc" )
	{
		presult = GetParams( str, i+1, param, 2, Com );

		if( presult && param[0] != "" && param[1] != "" )
		{
			glBlendFunc( GetConst( param[0], Com ), GetConst( param[1], Com ) );
			return;
		}
	}
	if( Com == "glFrontFace" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			glFrontFace( GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "gluQuadricDrawStyle" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			gluQuadricDrawStyle( obj, GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "gluQuadricNormals" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			gluQuadricNormals( obj, GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "gluQuadricOrientation" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			gluQuadricOrientation( obj, GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "gluQuadricTexture" )
	{
		presult = GetParams( str, i+1, param, 1, Com );

		if( presult && param[0] != "" )
		{
			gluQuadricTexture( obj, GetConst( param[0], Com ) );
			return;
		}
	}
	if( Com == "gluSphere" )
	{
		presult = GetParams( str, i+1, param, 3, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" )
		{
			gluSphere( obj, atof( param[0].c_str() ), atoi( param[1].c_str() ), atoi( param[2].c_str() ) );
			return;
		}
	}
	if( Com == "gluCylinder" )
	{
		presult = GetParams( str, i+1, param, 5, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" && param[3] != "" && param[4] != "" )
		{
			gluCylinder( obj, atof( param[0].c_str() ), atof( param[1].c_str() ), atof( param[2].c_str() ), atoi( param[3].c_str() ), atoi( param[4].c_str() ) );
			return;
		}
	}
	if( Com == "gluDisk" )
	{
		presult = GetParams( str, i+1, param, 4, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" && param[3] != "" )
		{
			gluDisk( obj, atof( param[0].c_str() ), atof( param[1].c_str() ), atoi( param[2].c_str() ), atoi( param[3].c_str() ) );
			return;
		}
	}
	if( Com == "gluPartialDisk" )
	{
		presult = GetParams( str, i+1, param, 6, Com );

		if( presult && param[0] != "" && param[1] != "" && param[2] != "" && param[3] != "" && param[4] != "" && param[5] != "" )
		{
			gluPartialDisk( obj, atof( param[0].c_str() ), atof( param[1].c_str() ), atoi( param[2].c_str() ), atoi( param[3].c_str() ), atof( param[4].c_str() ), atof( param[4].c_str() ) );
			return;
		}
	}

	Log.Error( "GLMODEL( " + file + " ): Bdne, lub z niewaciwymi parametrami polecenie: " + Com + ", oryginalny cig: " + str );
}

/*=====METODA CallObject=====
	Wywouje dan list wywietlania
*/
void GLModel::CallObject( unsigned int index )
{
	if( !loaded )
		return;

	if( index >= ListCount )
		return;

	glCallList( List + index );
}

/*=====METODA ReadHeader=====
	Czyta nagwek pliku GLM.
*/
bool GLModel::ReadHeader( FILE* fp )
{
	std::string str;

	// Pobieramy liczb tekstur
	str = GetString( fp );
	if( !sscanf( str.c_str(), "TEXCOUNT\t%u", &TexCount ) )
	{
		Log.Error( "GLMODEL( " + file + " ): Nie mona odczyta liczby tekstur!" );
		return false;
	}

	// Pobieramy liczb obiektw
	str = GetString( fp );
	if( !sscanf( str.c_str(), "LISTCOUNT\t%u", &ListCount ) )
	{
		Log.Error( "GLMODEL( " + file + " ): Nie mona odczyta liczby obiektw!" );
		return false;
	}

	// Pobieramy, czy plik ma zapisan animacj
	str = GetString( fp );
	if( !sscanf( str.c_str(), "ANIMATION\t%u", &Animation ) )
	{
		Log.Error( "GLMODEL( " + file + " ): Nie mona odczyta animacji!" );
		return false;
	}

	// Sprawdzamy czy to koniec nagwka
	str = GetString( fp );
	if( str != "END HEADER" )
	{
		Log.Error( "GLMODEL( " + file + " ): Brak koca nagwka!" );
		return false;
	}

	return true;
}

bool GLModel::LoadModel( std::string filename )
{
	// Sprawdzamy czy acuch nie jest pusty
	if( filename == "" )
	{
		Log.Error( "GLMODEL( " + file + " ): Pusty cig nazwy pliku!" );
		return false;
	}
	
	// zmienna tymczasowa, trzymajca jedn linie z pliku
	std::string str;

	// zmienna trzymajca numer wersji pliku
	int Version = 0;

	int i;

	// Prbujemy otworzy plik
	FILE* fp = fopen( filename.c_str(), "rt" );

	// Sprawdzamy poczenie
	if( !fp )
	{
		Log.Error( "GLMODEL( " + file + " ): plik " + filename + " nie istnieje lub cieka niewaciwa." );
		return false;
	}

	// Pobieramy pierwsz linie
	str = GetString( fp );

	// Skanujemy lini w poszukiwaniu numeru wersji
	if( !sscanf( str.c_str(), "GLM\t%d", &Version ) )
	{
		Log.Error( "GLMODEL( " + file + " ): Nieprawidowa pierwsza linia pliku " + filename + "!" );
		fclose( fp );
		return false;
	}

	// Sprawdzamy wersj
	if( Version > GLM_VERSION )
	{
		Log.Error( "GLMODEL( " + file + " ): Zbyt wysoka wersja pliku!" );
		fclose( fp );
		return false;
	}

	// Teraz sprawdzamy, czy ju jaki obiekt nie by zaadowany
	if( loaded )
	{
		Log.Report( "GLMODEL( " + file + " ): Przeadowanie pliku na " + filename );
		Free();
	}

	Log.Log( "GLMODEL( " + file + " ): adowanie modelu z pliku " + filename );
	file = filename;

	str = GetString( fp );

	// Czytamy nagwek
	if( str == "HEADER" )
	{
		if( !ReadHeader( fp ) )
		{
			Log.Error( "GLMODEL( " + file + " ): Bd w nagwku!" );
			fclose( fp );
			return false;
		}
	}
	else
	{
		Log.Error( "GLMODEL( " + file + " ): Brak nagwka!" );
		fclose( fp );
		return false;
	}

	if( Animation )
	{
		str = GetString( fp );

		// Czytamy nagwek
		if( str == "ANIMHEADER" )
		{
			if( !ReadAnimHeader( fp ) )
			{
				Log.Error( "GLMODEL( " + file + " ): Bd w nagwku animacji!" );
				fclose( fp );
				return false;
			}
		}
		else
		{
			Log.Error( "GLMODEL( " + file + " ): Brak nagwka animacji!" );
			fclose( fp );
			return false;
		}
	}
	
	// Czytamy tekstury
	if( TexCount > 0 )
	{
		Tex = new ioTexture[TexCount];

		str = GetString( fp );

		if( str == "TEXLIST" )
		{
			for( i = 0; i < TexCount; i++ )
			{
				str = GetString( fp );

				if( !Tex[i].LoadTGA( str ) )
				{
					Log.Error( "GLMODEL( " + file + " ): Bd przy adowaniu tekstury!" );
				}

				if( str == "END TEXLIST" )
				{
					Log.Error( "GLMODEL( " + file + " ): Przedwczesny koniec listy tekstur!" );
					break;
				}
			}
			
			if( str != "END TEXLIST" )
			{
				str = GetString( fp );
				if( str != "END TEXLIST" )
				{
					Log.Error( "GLMODEL( " + file + " ): Brak koca listy tekstur!" );
					Free();
					fclose( fp );
					return false;
				}
			}
		}
		else
		{
			Log.Error( "GLMODEL( " + file + " ): Brak listy tekstur!" );
			Free();
			fclose( fp );
			return false;
		}
	}

	// Czytamy modele
	if( ListCount > 0 )
	{
		List = glGenLists( ListCount );
		obj = gluNewQuadric();
		gluQuadricTexture( obj, GL_TRUE );

		for( i = 0; i < ListCount; i++ )
		{
			str = GetString( fp );

			int j;

			sscanf( str.c_str(), "MODEL\t%d", &j );
			
			if( i != j )
			{
				Log.Error( "GLMODEL( " + file + " ): Bd, za kolejno modeli!" );
				Free();
				fclose( fp );
				return false;
			}

			glNewList( List + i, GL_COMPILE );

			do
			{
				str = GetString( fp );

				if( str != "END MODEL" )
				{
					ParseGLCommand( str );
				}
			}
			while( str != "END MODEL" );

			glEndList();
		}

		gluDeleteQuadric( obj );
	}

	if( Animation )
	{
		if( FrameCount > 0 )
		{
			Frame = glGenLists( FrameCount );
			obj = gluNewQuadric();
			gluQuadricTexture( obj, GL_TRUE );

			for( i = 0; i < FrameCount; i++ )
			{
				str = GetString( fp );

				int j;

				sscanf( str.c_str(), "FRAME\t%d", &j );
				
				if( i != j )
				{
					Log.Error( "GLMODEL( " + file + " ): Bd, za kolejno klatek!" );
					Free();
					fclose( fp );
					return false;				
				}

				glNewList( Frame + i, GL_COMPILE );

				do
				{
					str = GetString( fp );

					if( str != "END FRAME" )
					{
						ParseGLCommand( str );
					}
				}
				while( str != "END FRAME" );

				glEndList();
			}

			gluDeleteQuadric( obj );
		}
	
	}

	str = GetString( fp );

	if( str == "END GLM" )
	{
		loaded = true;
		return true;
	}

	Log.Error( "GLMODEL( " + file + " ): Brak koca pliku!" );
	return false;
}

void GLModel::Free()
{
	if( Tex != NULL )
	{
		delete[] Tex;
		Tex = NULL;
		TexCount = 0;
	}

	if( glIsList( List ) )
	{
		glDeleteLists( List, ListCount );
		List = 0;
		ListCount = 0;
	}

	if( glIsList( Frame ) )
	{
		glDeleteLists( Frame, FrameCount );
		Frame = 0;
		FrameCount = 0;
	}

	loaded = false;
}


//===========METODY NIETESTOWANE!!!============
bool GLModel::ReadAnimHeader( FILE* fp )
{
	std::string str;

	// Pobieramy liczb tekstur
	str = GetString( fp );
	if( !sscanf( str.c_str(), "FRAMECOUNT\t%u", &FrameCount ) )
	{
		Log.Error( "GLMODEL( " + file + " ): Nie mona odczyta liczby klatek animacji!" );
		return false;
	}

	// Sprawdzamy czy to koniec nagwka
	str = GetString( fp );
	if( str != "END ANIMHEADER" )
	{
		Log.Error( "GLMODEL( " + file + " ): Brak koca nagwka animacji!" );
		return false;
	}

	return true;
}

void GLModel::PlayAnim( unsigned int fromframe, unsigned int toframe, bool canskip )
{
	if( !Animation )
		return;

	if( fromframe >= FrameCount || toframe < fromframe || toframe > FrameCount )
		return;

	if( playing && !canskip )
		return;

	FromFrame = fromframe;
	CurrFrame = fromframe;
	ToFrame = toframe;
	playing = true;
}

void GLModel::DoEngineAnim()
{
	if( !playing )
		return;

	CurrFrame++;

	if( CurrFrame > ToFrame )
		playing = false;
}

void GLModel::DoDrawAnim( unsigned int index )
{
	if( !Animation || !playing )
	{
		CallObject( index );
		return;
	}

	glCallList( Frame + CurrFrame );
}